@@ -10,7 +10,7 @@ import copy
 import numpy as np
 import scipy.sparse
 
-from hynet.types_ import hynet_int_, hynet_float_, hynet_sparse_
+from hynet.types_ import hynet_int_, hynet_float_, hynet_complex_, hynet_sparse_
 from hynet.utilities.worker import workers
 from hynet.utilities.base import (create_sparse_matrix,
                                   create_sparse_diag_matrix,
@@ -127,10 +127,10 @@ class ObjectiveFunction:
         y : .hynet_float_
             Function value ``y`` at the given point ``p``.
         """
-        return np.asscalar(p.v.conj().transpose().dot(self.C.dot(p.v)).real
-                           + self.c.transpose().dot(p.f)
-                           + self.a.transpose().dot(p.s)
-                           + self.r.transpose().dot(p.z))
+        return (p.v.conj().transpose().dot(self.C.dot(p.v)).real
+                + self.c.transpose().dot(p.f)
+                + self.a.transpose().dot(p.s)
+                + self.r.transpose().dot(p.z)).item()
 
     def __repr__(self):
         return ("ObjectiveFunction("
@@ -192,13 +192,13 @@ class Constraint(namedtuple('ConstraintBase', ['name', 'table', 'id',
         """
         result = -self.b
         if self.C is not None:
-            result += np.asscalar(p.v.conj().transpose().dot(self.C.dot(p.v)).real)
+            result += p.v.conj().transpose().dot(self.C.dot(p.v)).real.item()
         if self.c is not None:
-            result += np.asscalar(self.c.transpose().dot(p.f))
+            result += self.c.transpose().dot(p.f).item()
         if self.a is not None:
-            result += np.asscalar(self.a.transpose().dot(p.s))
+            result += self.a.transpose().dot(p.s).item()
         if self.r is not None:
-            result += np.asscalar(self.r.transpose().dot(p.z))
+            result += self.r.transpose().dot(p.z).item()
         return result
 
     def __repr__(self):
@@ -254,14 +254,16 @@ class QCQP:
     ub : QCQPPoint
         Specification of the upper bounds on ``|v|``, ``f``, ``s``, and ``z``.
     edges : (numpy.ndarray[.hynet_int_], numpy.ndarray[.hynet_int_])
-        Specification of the sparsity pattern of the constraint matrices via
-        the edges ``(edge_src, edge_dst)`` of an undirected graph. The constraint
-        matrices may only have nonzero entries in row/column ``edge_src[i]`` and
-        column/row ``edge_dst[i]``, with ``i in range(K)`` where
-        ``K = len(edge_src) = len(edge_dst)``, as well as on the diagonal. The
-        edges specified in ``edges`` must be *unique*, i.e., there must not exist
-        any parallel or anti-parallel edges. If case of any doubts, call
-        ``hynet.utilities.graph.eliminate_parallel_edges`` before assigning.
+        Specification of the sparsity pattern of the lower-triangular part of
+        the constraint matrices via the edges ``(edge_src, edge_dst)`` of an
+        undirected graph. The constraint matrices may only have nonzero entries
+        in row/column ``edge_src[i]`` and column/row ``edge_dst[i]``, with
+        ``i in range(K)`` where ``K = len(edge_src) = len(edge_dst)``, as well
+        as on the diagonal. The edges specified in ``edges`` must be *unique*,
+        i.e., there must not exist any parallel or anti-parallel edges, and
+        must refer to the *lower-triangular part*, i.e., ``edges[0] > edges[1]``.
+        If case of any doubts, call ``hynet.utilities.graph.eliminate_parallel_edges``
+        before assigning.
     roots : numpy.ndarray[.hynet_int_]
         Root nodes of the undirected graph defined by ``edges`` that specifies
         the sparsity pattern. At these root nodes the absolute angle of the
@@ -291,6 +293,9 @@ class QCQP:
     def __init__(self, obj_func, eq_crt, ineq_crt, lb, ub, edges, roots,
                  normalization=None, initial_point=None,
                  use_buffered_vectorization=True):
+        if not np.all(edges[0] > edges[1]):
+            raise RuntimeError("The edges of the sparsity graph must specify "
+                               "the pattern of the lower-triangular part.")
         self.obj_func = obj_func
         self.eq_crt = eq_crt
         self.ineq_crt = ineq_crt
@@ -389,7 +394,7 @@ class QCQP:
             [                                    ]         [   ]
             [ _mtx2vec(C_N), c_N^T, a_N^T, r_N^T ]         [b_N]
 
-            \____________________________________/         \___/
+           \\____________________________________/        \\___/
                              A                     x <=/==   b
 
         Parameters
@@ -648,17 +653,62 @@ def _vectorize_constraint(i, crt, aux):
             list_idx_col.append(idx_col)
             list_data.append(C_diag[idx_col])
         if aux.e_src.size:
-            C_offd = np.asarray(crt.C.tocsr()[aux.e_src, aux.e_dst])[0]
-            idx_col = C_offd.nonzero()[0]
+            # Extract and arrange the off-diagonal elements
+            #
+            # The sparsity graph (e_src, e_dst) refers to the lower-triangular
+            # part, which is extracted and vectorized according to the ordering
+            # of the edges of the sparsity graph. This can be performed with
+            #
+            # C_offd = np.asarray(crt.C.tocsr()[aux.e_src, aux.e_dst])[0]
+            # idx_col = C_offd.nonzero()[0]
+            # data_lt = C_offd[idx_col]
+            #
+            # To improve performance (and fix an issue with NumPy 1.17), the
+            # association with the sparsity graph edges is done manually below.
+            #
+            mask = (crt.C.row > crt.C.col)  # Lower-triangular off-diag. elements
+            idx_col = []
+            data_lt = []
+            for row, col, data in zip(crt.C.row[mask],
+                                      crt.C.col[mask],
+                                      crt.C.data[mask]):
+                if data == 0:
+                    continue
+                idx = np.where(aux.e_src == row)[0]
+                if not idx.size:
+                    raise RuntimeError("The sparsity pattern is violated.")
+                idx = idx[aux.e_dst[idx] == col]
+                if len(idx) != 1:
+                    raise RuntimeError("The sparsity pattern is violated.")
+                idx_col.append(idx[0])
+                data_lt.append(data)
+            idx_col = np.array(idx_col, dtype=hynet_int_)
+            data_lt = np.array(data_lt, dtype=hynet_complex_)
+            #
+            # ARCHIVAL CODE: Verification of the "manual" association
+            #
+            # Note that with NumPy 1.17, the indexing operation to construct
+            # C_offd causes the "ValueError: cannot set WRITEABLE flag to True
+            # of this array" when using parallel processing. For the verifi-
+            # cation, *hynet* must thus be run without parallel processing.
+            #
+            # C_offd = np.asarray(crt.C.tocsr()[aux.e_src, aux.e_dst])[0]
+            # idx = np.argsort(idx_col)
+            # idx_col = idx_col[idx]
+            # data_lt = data_lt[idx]
+            # if not np.all(idx_col == C_offd.nonzero()[0]):
+            #     raise RuntimeError()
+            # if not np.all(data_lt == C_offd[idx_col]):
+            #     raise RuntimeError()
+            #
             if idx_col.size:
                 idx_row = i + np.zeros(len(idx_col), dtype=hynet_int_)
-                data = C_offd[idx_col]
                 list_idx_row.append(idx_row)
                 list_idx_col.append(idx_col + aux.ofs_v_real)
-                list_data.append(2*data.real)
+                list_data.append(2*data_lt.real)
                 list_idx_row.append(idx_row)
                 list_idx_col.append(idx_col + aux.ofs_v_imag)
-                list_data.append(2*data.imag)
+                list_data.append(2*data_lt.imag)
     if crt.c is not None:
         # REMARK: This code includes the transposition by exchanging
         # the row and column index and it utilizes the fact that the
