@@ -198,13 +198,13 @@ class Scenario:
         Load a scenario from a *hynet* grid database.
     """
     def __init__(self):
-        self.id = 0
+        self.id = hynet_id_(0)
         self.name = ''
-        self.time = 0
+        self.time = hynet_float_(0)
         self.database_uri = ''
         self.grid_name = ''
-        self.base_mva = np.nan
-        self.loss_price = 0
+        self.base_mva = hynet_float_(np.nan)
+        self.loss_price = hynet_float_(0)
         self.description = ''
         self.annotation = ''
 
@@ -334,7 +334,7 @@ class Scenario:
         _log.debug("Scenario deep copy ({:.3f} sec.)".format(timer.total()))
         return scr
 
-    def verify(self):
+    def verify(self, log=_log.warning):
         """
         Verify the integrity and validity of the scenario.
 
@@ -343,15 +343,21 @@ class Scenario:
         data frames), proper (e.g., constraint limits), and valid (i.e.,
         compliant with all preconditions as assumed by *hynet*).
 
-        Besides errors being reported via an exception, critical settings are
-        reported by issuing a warning via the logging system.
+        Parameters
+        ----------
+        log : function(str) or None
+            Function to log information about critical settings (default
+            is the warning log of the module). Set to ``None`` to suppress this
+            log output.
 
         Raises
         ------
         ValueError
             In case any kind of integrity or validity violation is detected.
         """
-        verify_scenario(self)
+        timer = Timer()
+        verify_scenario(self, log)
+        _log.debug("Scenario verification ({:.3f} sec.)".format(timer.total()))
 
     @property
     def num_buses(self):
@@ -380,7 +386,7 @@ class Scenario:
 
         **Remark:** This property is *index-based* and intended for internal use.
         """
-        return pd.Series(self.get_bus_index(self.branch['src'].values),
+        return pd.Series(self.get_bus_index(self.branch['src'].to_numpy()),
                          index=self.branch.index)
 
     @property
@@ -390,7 +396,7 @@ class Scenario:
 
         **Remark:** This property is *index-based* and intended for internal use.
         """
-        return pd.Series(self.get_bus_index(self.branch['dst'].values),
+        return pd.Series(self.get_bus_index(self.branch['dst'].to_numpy()),
                          index=self.branch.index)
 
     @property
@@ -400,7 +406,7 @@ class Scenario:
 
         **Remark:** This property is *index-based* and intended for internal use.
         """
-        return pd.Series(self.get_bus_index(self.converter['src'].values),
+        return pd.Series(self.get_bus_index(self.converter['src'].to_numpy()),
                          index=self.converter.index)
 
     @property
@@ -410,7 +416,7 @@ class Scenario:
 
         **Remark:** This property is *index-based* and intended for internal use.
         """
-        return pd.Series(self.get_bus_index(self.converter['dst'].values),
+        return pd.Series(self.get_bus_index(self.converter['dst'].to_numpy()),
                          index=self.converter.index)
 
     @property
@@ -420,15 +426,17 @@ class Scenario:
 
         **Remark:** This property is *index-based* and intended for internal use.
         """
-        return pd.Series(self.get_bus_index(self.injector['bus'].values),
+        return pd.Series(self.get_bus_index(self.injector['bus'].to_numpy()),
                          index=self.injector.index)
 
     def get_bus_index(self, bus_id):
         """
-        Return the bus index for the given bus identifier.
+        Return the bus index(es) for the given (iterable of) bus identifier(s).
 
         **Remark:** This result is *index-based* and intended for internal use.
         """
+        if isinstance(bus_id, hynet_id_):
+            return hynet_int_(self.bus.index.get_loc(bus_id))
         return np.array([self.bus.index.get_loc(key) for key in bus_id],
                         dtype=hynet_int_)
 
@@ -438,17 +446,29 @@ class Scenario:
 
         **Remark:** This result is *index-based* and intended for internal use.
         """
-        return np.nonzero(self.bus['ref'].values)[0]
+        return np.nonzero(self.bus['ref'].to_numpy())[0]
 
-    def has_hybrid_architecture(self):
+    def verify_hybrid_architecture_conditions(self, log=_log.warning):
         """
-        Return ``True`` if the scenario features the *hybrid architecture*.
+        Return ``True`` if the *hybrid architecture*'s exactness results hold.
 
-        The *hybrid architecture* denotes a particular class of power systems
-        that, under very mild conditions, induces exactness to the semidefinite
+        The *hybrid architecture* denotes a class of network topologies that,
+        under very mild conditions, induces exactness to the semidefinite
         and second-order cone relaxation of the OPF problem in case that no
         *pathological price profile* emerges, see [1]_, [2]_, and [3]_. This
-        method assumes validity of the scenario, see ``Scenario.verify``.
+        function returns ``True`` if the topological requirements of the
+        *hybrid architecture* as well as the conditions on the system
+        parameters that are utilized for the aforementioned results on
+        exactness are satisfied for this scenario. It is assumed that the
+        validity of the scenario is established beforehand, see
+        ``Scenario.verify``.
+
+        Parameters
+        ----------
+        log : function(str) or None
+            Function to log information about violated conditions (default
+            is the warning log of the module). Set to ``None`` to suppress any
+            log output.
 
         References
         ----------
@@ -458,63 +478,63 @@ class Scenario:
         .. [2] M. Hotz and W. Utschick, "The Hybrid Transmission Grid
                Architecture: Benefits in Nodal Pricing," in IEEE Trans. Power
                Systems, vol. 33, no. 2, pp. 1431-1442, Mar. 2018.
-        .. [3] M. Hotz and W. Utschick, "hynet: An optimal power flow framework
-               for hybrid AC/DC power systems," arXiv:1811.10496, Nov. 2018.
-               [Online]. Available: http://arxiv.org/abs/1811.10496
+        .. [3] M. Hotz and W. Utschick, "hynet: An Optimal Power Flow Framework
+               for Hybrid AC/DC Power Systems," in IEEE Trans. Power Systems,
+               vol. 35, no. 2, pp. 1036-1047, Mar. 2020.
 
         See Also
         --------
         hynet.scenario.representation.Scenario.verify
         """
-        return verify_hybrid_architecture(self)
+        return verify_hybrid_architecture(self, log)
 
     def has_acyclic_subgrids(self):
         """
         Return ``True`` if all subgrids are acyclic and ``False`` otherwise.
         """
         return is_acyclic_graph(np.arange(self.num_buses),
-                                (self.e_src.values, self.e_dst.values))
+                                (self.e_src.to_numpy(), self.e_dst.to_numpy()))
 
     def has_acyclic_ac_subgrids(self):
         """
         Return ``True`` if all AC subgrids are acyclic and ``False`` otherwise.
         """
-        ac_buses = np.where(np.equal(self.bus['type'].values, BusType.AC))[0]
-        mask = np.logical_and(np.isin(self.e_src.values, ac_buses),
-                              np.isin(self.e_dst.values, ac_buses))
-        return is_acyclic_graph(ac_buses, (self.e_src.values[mask],
-                                           self.e_dst.values[mask]))
+        ac_buses = np.where(np.equal(self.bus['type'].to_numpy(), BusType.AC))[0]
+        mask = np.logical_and(np.isin(self.e_src.to_numpy(), ac_buses),
+                              np.isin(self.e_dst.to_numpy(), ac_buses))
+        return is_acyclic_graph(ac_buses, (self.e_src.to_numpy()[mask],
+                                           self.e_dst.to_numpy()[mask]))
 
     def has_acyclic_dc_subgrids(self):
         """
         Return ``True`` if all DC subgrids are acyclic and ``False`` otherwise.
         """
-        dc_buses = np.where(np.equal(self.bus['type'].values, BusType.DC))[0]
-        mask = np.logical_and(np.isin(self.e_src.values, dc_buses),
-                              np.isin(self.e_dst.values, dc_buses))
-        return is_acyclic_graph(dc_buses, (self.e_src.values[mask],
-                                           self.e_dst.values[mask]))
+        dc_buses = np.where(np.equal(self.bus['type'].to_numpy(), BusType.DC))[0]
+        mask = np.logical_and(np.isin(self.e_src.to_numpy(), dc_buses),
+                              np.isin(self.e_dst.to_numpy(), dc_buses))
+        return is_acyclic_graph(dc_buses, (self.e_src.to_numpy()[mask],
+                                           self.e_dst.to_numpy()[mask]))
 
     def get_ac_subgrids(self):
         """
         Return a list with a pandas index of bus IDs for every AC subgrid.
         """
-        ac_buses = np.where(np.equal(self.bus['type'].values, BusType.AC))[0]
-        mask = np.logical_and(np.isin(self.e_src.values, ac_buses),
-                              np.isin(self.e_dst.values, ac_buses))
-        subgrids = get_graph_components(ac_buses, (self.e_src.values[mask],
-                                                   self.e_dst.values[mask]))
+        ac_buses = np.where(np.equal(self.bus['type'].to_numpy(), BusType.AC))[0]
+        mask = np.logical_and(np.isin(self.e_src.to_numpy(), ac_buses),
+                              np.isin(self.e_dst.to_numpy(), ac_buses))
+        subgrids = get_graph_components(ac_buses, (self.e_src.to_numpy()[mask],
+                                                   self.e_dst.to_numpy()[mask]))
         return [self.bus.index[subgrid] for subgrid in subgrids]
 
     def get_dc_subgrids(self):
         """
         Return a list with a pandas index of bus IDs for every DC subgrid.
         """
-        dc_buses = np.where(np.equal(self.bus['type'].values, BusType.DC))[0]
-        mask = np.logical_and(np.isin(self.e_src.values, dc_buses),
-                              np.isin(self.e_dst.values, dc_buses))
-        subgrids = get_graph_components(dc_buses, (self.e_src.values[mask],
-                                                   self.e_dst.values[mask]))
+        dc_buses = np.where(np.equal(self.bus['type'].to_numpy(), BusType.DC))[0]
+        mask = np.logical_and(np.isin(self.e_src.to_numpy(), dc_buses),
+                              np.isin(self.e_dst.to_numpy(), dc_buses))
+        subgrids = get_graph_components(dc_buses, (self.e_src.to_numpy()[mask],
+                                                   self.e_dst.to_numpy()[mask]))
         return [self.bus.index[subgrid] for subgrid in subgrids]
 
     def get_islands(self):
@@ -522,10 +542,10 @@ class Scenario:
         Return a list with a pandas index of bus IDs for every islanded grid.
         """
         islands = get_graph_components(np.arange(self.num_buses),
-                                       (np.concatenate((self.e_src.values,
-                                                        self.c_src.values)),
-                                        np.concatenate((self.e_dst.values,
-                                                        self.c_dst.values))))
+                                       (np.concatenate((self.e_src.to_numpy(),
+                                                        self.c_src.to_numpy())),
+                                        np.concatenate((self.e_dst.to_numpy(),
+                                                        self.c_dst.to_numpy()))))
         return [self.bus.index[island] for island in islands]
 
     def analyze_cycles(self):
@@ -570,8 +590,8 @@ class Scenario:
 
         for n, subgrid in enumerate(subgrids):
             idx = src.isin(subgrid) | dst.isin(subgrid)
-            num_corridors = len(eliminate_parallel_edges((src.loc[idx].values,
-                                                          dst.loc[idx].values))[0])
+            num_corridors = len(eliminate_parallel_edges((src.loc[idx].to_numpy(),
+                                                          dst.loc[idx].to_numpy()))[0])
             result.loc[n + 1] = (num_corridors - (len(subgrid) - 1),
                                  self.bus.at[subgrid[0], 'type'],
                                  subgrid)
@@ -598,7 +618,7 @@ class Scenario:
         if self.branch.empty or not corridors[0].size:
             return pd.Index([], dtype=hynet_id_, name=self.branch.index.name)
 
-        edges = (self.e_src.values, self.e_dst.values)
+        edges = (self.e_src.to_numpy(), self.e_dst.to_numpy())
         E = np.column_stack((edges[0], edges[1]))
         E.sort(axis=1)
 
@@ -618,7 +638,7 @@ class Scenario:
         if self.branch.empty:
             return parallel_branches
 
-        E = np.column_stack((self.e_src.values, self.e_dst.values))
+        E = np.column_stack((self.e_src.to_numpy(), self.e_dst.to_numpy()))
         E.sort(axis=1)
 
         (C, count) = np.unique(E, axis=0, return_counts=True)
@@ -639,14 +659,14 @@ class Scenario:
         """
         Return a pandas index with the branch IDs of all AC branches.
         """
-        mask = self.bus.loc[self.branch['src'], 'type'].values == BusType.AC
+        mask = self.bus.loc[self.branch['src'], 'type'].to_numpy() == BusType.AC
         return self.branch.index[mask]
 
     def get_dc_branches(self):
         """
         Return a pandas index with the branch IDs of all DC branches.
         """
-        mask = self.bus.loc[self.branch['src'], 'type'].values == BusType.DC
+        mask = self.bus.loc[self.branch['src'], 'type'].to_numpy() == BusType.DC
         return self.branch.index[mask]
 
     def remove_buses(self, buses):
@@ -702,6 +722,17 @@ class Scenario:
         return "{0.days}d {0.hours:02d}:{0.minutes:02d}:{0.seconds:02d}"\
                .format(self.get_time_tuple())
 
+    def get_relative_loading(self):
+        """
+        Return the ratio of tot. active power load to tot. active power inj. cap.
+
+        This ratio indicates the relative amount of the active power injection
+        capacity that is utilized by the active power load of this scenario.
+        """
+        p_capacity = sum(x.p_max for x in self.injector['cap'])
+        p_tot_load = self.bus['load'].to_numpy().real.sum()
+        return p_tot_load/p_capacity
+
     def set_conservative_rating(self):
         """
         Adjust the branch rating to a conservative setting (MATPOWER compat.).
@@ -722,28 +753,37 @@ class Scenario:
                Trans. Power Systems, vol. 31, no. 6, pp. 4504-4516, Nov. 2016.
         """
         v_max = self.bus['v_max']
-        v_max_branch = np.maximum(v_max.loc[self.branch['src']].values,
-                                  v_max.loc[self.branch['dst']].values)
+        v_max_branch = np.maximum(v_max.loc[self.branch['src']].to_numpy(),
+                                  v_max.loc[self.branch['dst']].to_numpy())
         self.branch['rating'] /= v_max_branch
 
-    def set_minimum_series_resistance(self, min_resistance):
+    def set_minimum_series_resistance(self, min_resistance, branch_ids=None):
         """
-        Set the series resistance of all branches to a minimum of ``min_resistance``.
+        Set the series resistance of the branches to a minimum of ``min_resistance``.
 
         Very small values of the series resistance of branches may lead to
         numerical issues during the solution of the OPF problem. With this
-        method, the series resistance is enforced to be ``>= min_resistance``.
+        method, the series resistance is enforced to be larger or equal to
+        ``min_resistance``.
 
         Parameters
         ----------
         min_resistance : float
-            Minimum resistance for the series resistance of all branches. If a
-            branch has a series resistance below this value, it is replaced by
-            ``min_resistance``.
+            Minimum resistance for the series resistance of the specified
+            branches. If a branch has a series resistance below this value,
+            it is replaced by ``min_resistance``.
+        branch_ids : Iterable[.hynet_id_]
+            Iterable of branch IDs for which the minimum series resistance
+            shall be ensured. By default, all branches are considered.
         """
-        idx = self.branch['z_bar'].real < min_resistance
-        self.branch.loc[idx, 'z_bar'] =\
-            min_resistance + 1j*self.branch.loc[idx, 'z_bar'].imag
+        if branch_ids is None:
+            branch_ids = self.branch.index
+
+        z_bar = self.branch.loc[branch_ids, 'z_bar']
+        affected = z_bar.index[z_bar.to_numpy().real < min_resistance]
+
+        self.branch.loc[affected, 'z_bar'] = \
+            min_resistance + 1j*z_bar.loc[affected].to_numpy().imag
 
     def ensure_reference(self):
         """
@@ -769,6 +809,171 @@ class Scenario:
                     p_max = cap.p_max
             self.bus.at[ref_bus, 'ref'] = True
 
+    def add_bus(self, type_, base_kv, v_min, v_max, ref=False, y_tld=0, load=0,
+                zone=None, annotation='', bus_id=None):
+        """
+        Add a bus to this scenario.
+
+        Please refer to the documentation of the ``bus`` data frame for a
+        description of the parameters. If ``bus_id`` is provided, this ID is
+        used for the bus, otherwise an appropriate ID is generated.
+
+        Returns
+        -------
+        bus_id : .hynet_id_
+            ID of the added bus.
+        """
+        if bus_id is None:
+            if self.bus.empty:
+                bus_id = 1
+            else:
+                bus_id = self.bus.index.max() + 1
+        bus_id = hynet_id_(bus_id)
+
+        bus = pd.DataFrame({
+            'type': [type_],
+            'ref': [bool(ref)],
+            'base_kv': [hynet_float_(base_kv)],
+            'y_tld': [hynet_complex_(y_tld)],
+            'load': [hynet_complex_(load)],
+            'v_min': [hynet_float_(v_min)],
+            'v_max': [hynet_float_(v_max)],
+            'zone': [zone if zone is None else hynet_id_(zone)],
+            'annotation': [annotation]
+        }, index=[bus_id])
+        bus.index.name = self.bus.index.name
+
+        self.bus = self.bus.append(bus, sort=False)
+        return bus_id
+
+    def add_branch(self, type_, src, dst, z_bar, y_src=0.0, y_dst=0.0,
+                   rho_src=1.0, rho_dst=1.0, length=np.nan, rating=np.nan,
+                   angle_min=np.nan, angle_max=np.nan, drop_min=np.nan,
+                   drop_max=np.nan, annotation='', branch_id=None):
+        """
+        Add a branch to this scenario.
+
+        Please refer to the documentation of the ``branch`` data frame for a
+        description of the parameters. If ``branch_id`` is provided, this ID is
+        used for the branch, otherwise an appropriate ID is generated.
+
+        Returns
+        -------
+        branch_id : .hynet_id_
+            ID of the added branch.
+        """
+        if branch_id is None:
+            if self.branch.empty:
+                branch_id = 1
+            else:
+                branch_id = self.branch.index.max() + 1
+        branch_id = hynet_id_(branch_id)
+
+        branch = pd.DataFrame({
+            'type': [type_],
+            'src': [hynet_id_(src)],
+            'dst': [hynet_id_(dst)],
+            'z_bar': [hynet_complex_(z_bar)],
+            'y_src': [hynet_complex_(y_src)],
+            'y_dst': [hynet_complex_(y_dst)],
+            'rho_src': [hynet_complex_(rho_src)],
+            'rho_dst': [hynet_complex_(rho_dst)],
+            'length': [hynet_float_(length)],
+            'rating': [hynet_float_(rating)],
+            'angle_min': [hynet_float_(angle_min)],
+            'angle_max': [hynet_float_(angle_max)],
+            'drop_min': [hynet_float_(drop_min)],
+            'drop_max': [hynet_float_(drop_max)],
+            'annotation': [annotation]
+        }, index=[branch_id])
+        branch.index.name = self.branch.index.name
+
+        self.branch = self.branch.append(branch, sort=False)
+        return branch_id
+
+    def add_converter(self, src, dst, cap_src, cap_dst, loss_fwd=0, loss_bwd=0,
+                      loss_fix=0, annotation='', converter_id=None):
+        """
+        Add a converter to this scenario.
+
+        Please refer to the documentation of the ``converter`` data frame
+        for a description of the parameters. If ``converter_id`` is provided,
+        this ID is used for the converter, otherwise an appropriate ID is
+        generated.
+
+        Returns
+        -------
+        converter_id : .hynet_id_
+            ID of the added converter.
+        """
+        if converter_id is None:
+            if self.converter.empty:
+                converter_id = 1
+            else:
+                converter_id = self.converter.index.max() + 1
+        converter_id = hynet_id_(converter_id)
+
+        converter = pd.DataFrame({
+            'src': [hynet_id_(src)],
+            'dst': [hynet_id_(dst)],
+            'cap_src': [cap_src],
+            'cap_dst': [cap_dst],
+            'loss_fwd': [hynet_float_(loss_fwd)],
+            'loss_bwd': [hynet_float_(loss_bwd)],
+            'loss_fix': [hynet_float_(loss_fix)],
+            'annotation': [annotation]
+        }, index=[converter_id])
+        converter.index.name = self.converter.index.name
+
+        self.converter = self.converter.append(converter, sort=False)
+        return converter_id
+
+    def add_injector(self, type_, bus, cap, cost_p=None, cost_q=None,
+                     cost_start=0, cost_stop=0, ramp_up=np.nan,
+                     ramp_down=np.nan, min_up=np.nan, min_down=np.nan,
+                     energy_min=np.nan, energy_max=np.nan, annotation='',
+                     injector_id=None):
+        """
+        Add an injector to this scenario.
+
+        Please refer to the documentation of the ``injector`` data frame
+        for a description of the parameters. If ``injector_id`` is provided,
+        this ID is used for the injector, otherwise an appropriate ID is
+        generated.
+
+        Returns
+        -------
+        injector_id : .hynet_id_
+            ID of the added injector.
+        """
+        if injector_id is None:
+            if self.injector.empty:
+                injector_id = 1
+            else:
+                injector_id = self.injector.index.max() + 1
+        injector_id = hynet_id_(injector_id)
+
+        injector = pd.DataFrame({
+            'type': [type_],
+            'bus': [hynet_id_(bus)],
+            'cap': [cap],
+            'cost_p': [cost_p],
+            'cost_q': [cost_q],
+            'cost_start': [hynet_float_(cost_start)],
+            'cost_stop': [hynet_float_(cost_stop)],
+            'ramp_up': [hynet_float_(ramp_up)],
+            'ramp_down': [hynet_float_(ramp_down)],
+            'min_up': [hynet_float_(min_up)],
+            'min_down': [hynet_float_(min_down)],
+            'energy_min': [hynet_float_(energy_min)],
+            'energy_max': [hynet_float_(energy_max)],
+            'annotation': [annotation]
+        }, index=[injector_id])
+        injector.index.name = self.injector.index.name
+
+        self.injector = self.injector.append(injector, sort=False)
+        return injector_id
+
     def add_compensator(self, bus, q_max, q_min=None, cost_q=None):
         """
         Add a reactive power compensator to this scenario.
@@ -794,30 +999,11 @@ class Scenario:
         if q_min is None:
             q_min = -q_max
 
-        if self.injector.empty:
-            injector_id = 1
-        else:
-            injector_id = self.injector.index.max() + 1
-
-        compensator = pd.DataFrame({
-            'type': [InjectorType.COMPENSATION],
-            'bus': [hynet_id_(bus)],
-            'cap': [CapRegion(p_bnd=(0, 0), q_bnd=(q_min, q_max))],
-            'cost_p': [None],
-            'cost_q': [cost_q],
-            'cost_start': [hynet_float_(0.0)],
-            'cost_stop': [hynet_float_(0.0)],
-            'ramp_up': [hynet_float_(np.nan)],
-            'ramp_down': [hynet_float_(np.nan)],
-            'min_up': [hynet_float_(np.nan)],
-            'min_down': [hynet_float_(np.nan)],
-            'energy_min': [hynet_float_(np.nan)],
-            'energy_max': [hynet_float_(np.nan)],
-            'annotation': ['']
-        }, index=[injector_id])
-
-        self.injector = self.injector.append(compensator, sort=False)
-        return injector_id
+        return self.add_injector(type_=InjectorType.COMPENSATION,
+                                 bus=bus,
+                                 cap=CapRegion(p_bnd=(0, 0),
+                                               q_bnd=(q_min, q_max)),
+                                 cost_q=cost_q)
 
     def __repr__(self):
         return ("\n+" + "-"*78 + "+\n"
