@@ -17,12 +17,17 @@ except ImportError:
     pass
 
 try:
-    from hynet.solver import mosek
-    from hynet.utilities.base import validate_mosek_license
+    from hynet.solver import cplex
 
-    if validate_mosek_license():
-        AVAILABLE_SOLVERS.append(mosek.SOCRSolver)
-        AVAILABLE_SOLVERS.append(mosek.SDRSolver)
+    AVAILABLE_SOLVERS.append(cplex.SOCRSolver)
+except ImportError:
+    pass
+
+try:
+    from hynet.solver import mosek
+
+    AVAILABLE_SOLVERS.append(mosek.SOCRSolver)
+    AVAILABLE_SOLVERS.append(mosek.SDRSolver)
 except ImportError:
     pass
 
