@@ -7,7 +7,7 @@ import logging
 
 import numpy as np
 import scipy.sparse
-import ipopt
+import cyipopt
 
 from hynet.types_ import (hynet_int_,
                           hynet_float_,
@@ -118,7 +118,7 @@ class QCQPSolver(SolverInterface):
         return result
 
 
-class IPOPTProblem(ipopt.problem):
+class IPOPTProblem(cyipopt.Problem):
     """
     Specification and solution of the QCQP via IPOPT.
     """
@@ -251,13 +251,13 @@ class IPOPTProblem(ipopt.problem):
                          cu=crt_ub)
 
         # Set options
-        self.addOption(b'sb', b'yes')  # Suppress the IPOPT banner
-        self.addOption(b'print_level', 5 if options.verbose else 0)
-        # self.addOption(b'print_options_documentation', b'yes')
+        self.add_option(b'sb', b'yes')  # Suppress the IPOPT banner
+        self.add_option(b'print_level', 5 if options.verbose else 0)
+        # self.add_option(b'print_options_documentation', b'yes')
         for key, value in options.param.items():
-            self.addOption(key.encode(encoding='utf-8'),
-                           value.encode(encoding='utf-8')
-                           if isinstance(value, str) else value)
+            self.add_option(key.encode(encoding='utf-8'),
+                            value.encode(encoding='utf-8')
+                            if isinstance(value, str) else value)
 
     def _split(self, x):
         """
